/*
 *
 * Copyright 2011 Texas Instruments, Inc.
 * Author: Archit Taneja <archit@ti.com>
 *
 * based on d2l panel driver by Jerry Alexander <x0135174@ti.com>
 *
 * This program iss free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details. 
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/debugfs.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/seq_file.h>
#include <linux/backlight.h>
#include <linux/fb.h>
#include <linux/interrupt.h>
#include <linux/gpio.h>
#include <linux/slab.h>
#include <linux/regulator/consumer.h>
#include <linux/mutex.h>
#include <linux/i2c.h>
#include <linux/of_device.h>
//#include <video/omapdss.h>
#include <video/mipi_display.h>
//#include <omap-panel-dsi85.h>

#include <drm/drm_crtc_helper.h>
#include <drm/drm_mipi_dsi.h>


#include <drm/drmP.h>
#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_edid.h>

//#include "panel-dsi85.h"

#define DSI85DEBUG

static struct i2c_board_info dsi85_i2c_board_info = {
	I2C_BOARD_INFO("dsi85_i2c_driver", 0x2d /*0x2c*/),
};

struct dsi85_i2c_data {
	struct mutex xfer_lock;
	struct device_node *host_node;
	u8 channel_id;
};
static struct i2c_client *g_dsi85_client = NULL;
static const struct i2c_device_id dsi85_i2c_id[] = {
	{ "dsi85_i2c_driver", 0 },
	{},
};

MODULE_DEVICE_TABLE(i2c, dsi85_i2c_id);

static struct of_device_id dsi85_dt_ids[] = {
	{ .compatible = "ti,sn65dsi8x", .data = NULL, },
	{ /* sentinel */ }
};

#define LVDS_CLK_FROM_DSI_CLK  1

struct dsi85_lvds_timings {
    u16 hfp;
    u16 hsw;
    u16 hbp;
    u16 vfp;
    u16 vsw;
    u16 vbp;
};

static struct dsi85_lvds_timings lvds_timings = {
	.hfp = 40,
	.hsw = 128,
	.hbp = 40,
	.vfp = 1,
	.vsw = 4,
	.vbp = 9,    
};

struct dsi85_reg {
	/* Address and register value */
	u8 data[10];
	int len;
};

static void dsi85_dumpconfig(struct i2c_client *client)
{
	struct i2c_client *dsi85_i2c_client = client;
	printk(KERN_INFO ">>>>>>>>>>>>>>>>>> luis' message\n");
    int val;
    val = i2c_smbus_read_byte_data(dsi85_i2c_client, 0x00);
    printk("@@@@@@@@@@@@@@@@@@@ val=%d\n", val);
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x00, i2c_smbus_read_byte_data(dsi85_i2c_client, 0x00));

	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x01, \
	i2c_smbus_read_byte_data(dsi85_i2c_client, 0x01));

	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x02, \
	i2c_smbus_read_byte_data(dsi85_i2c_client, 0x02));
	
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x03, \
	i2c_smbus_read_byte_data(dsi85_i2c_client, 0x03));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x04, \
	i2c_smbus_read_byte_data(dsi85_i2c_client, 0x04));
	
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x05, \
	i2c_smbus_read_byte_data(dsi85_i2c_client, 0x05));
	

	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x0D, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x0D));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x0A, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x0A));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x0B, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x0B));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x10, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x10));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x18, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x18));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x1A, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x1A));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x20, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x20));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x21, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x21));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x22, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x22));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x23, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x23));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x24, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x24));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x25, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x25));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x26, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x26));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x27, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x27));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x28, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x28));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x29, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x29));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2A, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2A));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2B, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2B));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2C, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2C));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2D, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2D));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2E, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2E));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x2F, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x2F));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x30, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x30));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x31, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x31));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x32, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x32));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x33, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x33));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x34, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x34));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x35, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x35));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x36, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x32));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x37, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x33));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x38, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x34));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x39, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x35));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x3A, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x35));
	printk(KERN_INFO "DSI85-I2C: read 0x%02x  - 0x%02x\n", 0x3B, \
		i2c_smbus_read_byte_data(dsi85_i2c_client, 0x3B));
}
static int dsi85_i2c_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	struct dsi85_i2c_data *dsi85;
	struct device *dev = &client->dev;

	printk(KERN_INFO ">>>>>>>>>>>>>>>>>>>>>>> DSI85-I2C: Probe called!\n");
	dsi85 = kzalloc(sizeof(struct dsi85_i2c_data), GFP_KERNEL);

	if (!dsi85) {
		printk(KERN_ERR "DSI85-I2C: memory allocation failed!\n");
		return -ENOMEM;
	}
    g_dsi85_client = client;
	mutex_init(&dsi85->xfer_lock);
	i2c_set_clientdata(client, dsi85);

   dsi85_dumpconfig(client);

	return 0;
}

static int dsi85_i2c_remove(struct i2c_client *client)
{
	struct dsi85_i2c_data *dsi85_i2c_data =
					i2c_get_clientdata(client);
	kfree(dsi85_i2c_data);
	return 0;
}

static struct i2c_driver dsi85_i2c_driver = {
	.driver = {
		.name	= "dsi85_i2c_driver",
		.owner = THIS_MODULE,
		.of_match_table = dsi85_dt_ids,
	},
	.probe		= dsi85_i2c_probe,
	.remove		= dsi85_i2c_remove,
	.id_table	= dsi85_i2c_id,
};




#if 0
static void dsi85_test(struct omap_dss_device *dssdev)
{
	printk(KERN_INFO "Now Configuring Test Pattern\n");
	
	struct dsi85_data *dsi85_d = dev_get_drvdata(&dssdev->dev);
	struct i2c_client *dsi85_i2c_client = dsi85_d->dsi85_i2c_client;
    u8 val = 0;

	/* Soft reset and disable PLL */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_SOFT_RESET, 0x01);
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_PLL_EN, 0x00);

#if LVDS_CLK_FROM_DSI_CLK
    val = 0x1;
#endif

	/* user external clock reference with no muliplier */
    if (dssdev->panel.timings.pixel_clock <= 37500)
    {
        // Do nothing.
    }
    else if (dssdev->panel.timings.pixel_clock <= 62500)
    {
        val |= (0x01 << 1);
    }
    else if (dssdev->panel.timings.pixel_clock <= 87500)
    {
        val |= (0x02 << 1);
    }
    else if (dssdev->panel.timings.pixel_clock <= 112500)
    {
        val |= (0x03 << 1);
    }
    else if (dssdev->panel.timings.pixel_clock <= 137500)
    {
        val |= (0x04 << 1);
    }
    else
    {
        val |= (0x05 << 1);
    }
    
    //LADLDL
    printk(KERN_INFO "Pixel CLK value was %d\n",val);
    
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CORE_PLL, val);
#if LVDS_CLK_FROM_DSI_CLK
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_PLL_DIV, 0x10);  // Divide DSI_CLK by 3.
#else
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_PLL_DIV, 0x00);  // Multiply REFCLK by 1.
#endif


	//LADLDL PLL enable after address A and B configured
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_PLL_EN, 0x01);



	/* four DSI lanes with single channel*/
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_DSI_CFG, 0x20);
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_DSI_EQ, 0x00);

    /* set DSI clock range */
    i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_DSI_CLK_RNG, (dssdev->panel.timings.pixel_clock * 3 / 5000));
    i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_DSI_CLK_RNG, (dssdev->panel.timings.pixel_clock * 3 / 5000));

	/* set LVDS for single channel, 24 bit mode, HS/VS low, DE high */
	//i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_LVDS_MODE, 0x7F);
	/*LADLD set LVDS for single channel, 24 bit mode, HS/VS low, DE high */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_LVDS_MODE, 0x60);

	/* set LVDS 200 Ohm termination and max differential swing voltage */
	//LADLDL
	//i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_LVDS_SIGN, 0x00);
	//i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_LVDS_TERM, 0x00);

	/* x resolution high/low for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_LINE_LEN_LO, \
		((dssdev->panel.timings.x_res) & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_LINE_LEN_HI, \
		((dssdev->panel.timings.x_res) & 0xFF00)>>8);

	/* x resolution high/low for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_LINE_LEN_LO, \
		(dssdev->panel.timings.x_res & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_LINE_LEN_HI, \
		(dssdev->panel.timings.x_res & 0xFF00)>>8);

	/* y resolution high/low for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VERT_LINES_LO, \
		(dssdev->panel.timings.y_res & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VERT_LINES_HI, \
		(dssdev->panel.timings.y_res & 0xFF00)>>8);

	/* y resolution high/low for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VERT_LINES_LO, \
		(dssdev->panel.timings.y_res & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VERT_LINES_HI, \
		(dssdev->panel.timings.y_res & 0xFF00)>>8);

	/* SYNC delay high/low for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, \
		DSI85_CHA_SYNC_DELAY_LO, 0x00);
	i2c_smbus_write_byte_data(dsi85_i2c_client, \
		DSI85_CHA_SYNC_DELAY_HI, 0x02);

	/* SYNC delay high/low for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, \
		DSI85_CHB_SYNC_DELAY_LO, 0x00);
	i2c_smbus_write_byte_data(dsi85_i2c_client, \
		DSI85_CHB_SYNC_DELAY_HI, 0x02);

	/* HSYNC width high/low for channel A */
/*	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HSYNC_WIDTH_LO, \
		(dssdev->panel.timings.hsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HSYNC_WIDTH_HI, \
		(dssdev->panel.timings.hsw & 0xFF00)>>8);	*/
    i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HSYNC_WIDTH_LO, \
		(lvds_timings.hsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HSYNC_WIDTH_HI, \
		(lvds_timings.hsw & 0xFF00)>>8);

	/* HSYNC width high/low for channel B */
/*	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HSYNC_WIDTH_LO, \
		(dssdev->panel.timings.hsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HSYNC_WIDTH_HI, \
		(dssdev->panel.timings.hsw & 0xFF00)>>8); */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HSYNC_WIDTH_LO, \
		(lvds_timings.hsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HSYNC_WIDTH_HI, \
		(lvds_timings.hsw & 0xFF00)>>8);

	/* VSYNC width high/low for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VSYNC_WIDTH_LO, \
		(lvds_timings.vsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VSYNC_WIDTH_HI, \
		(lvds_timings.vsw & 0xFF00)>>8);

	/* VSYNC width high/low for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VSYNC_WIDTH_LO, \
		(lvds_timings.vsw & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VSYNC_WIDTH_HI, \
		(lvds_timings.vsw & 0xFF00)>>8);

	/* Horizontal BackPorch for channel A */
//	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HORZ_BACKPORCH, \
//		(dssdev->panel.timings.hbp & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HORZ_BACKPORCH, \
		(lvds_timings.hbp & 0x00FF));

	/* Horizontal BackPorch for channel B */
//	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HORZ_BACKPORCH, \
//  	(dssdev->panel.timings.hbp & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HORZ_BACKPORCH, \
		(lvds_timings.hbp & 0x00FF));

	/* Vertical BackPorch for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VERT_BACKPORCH, \
		(lvds_timings.vbp & 0x00FF)); 

	/* Vertical BackPorch for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VERT_BACKPORCH, \
		(lvds_timings.vbp & 0x00FF));

	/* Horizontal FrontPorch for channel A */
//	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HORZ_FRONTPORCH, \
//		(dssdev->panel.timings.hfp & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_HORZ_FRONTPORCH, \
		(lvds_timings.hfp & 0x00FF));

	/* Horizontal FrontPorch for channel B */
//	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HORZ_FRONTPORCH, \
//		(dssdev->panel.timings.hfp & 0x00FF));
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_HORZ_FRONTPORCH, \
		(lvds_timings.hfp & 0x00FF));

	/* Vertical FrontPorch for channel A */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHA_VERT_FRONTPORCH, \
		(lvds_timings.vbp & 0x00FF));

	/* Vertical FrontPorch for channel B */
	i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_CHB_VERT_FRONTPORCH, \
		(lvds_timings.vbp & 0x00FF));

	//Test Pattern
	i2c_smbus_write_byte_data(dsi85_i2c_client, 0x3C, 0x11);

	//LADLDL
	/* Soft reset and enable PLL */
	//i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_SOFT_RESET, 0x01);
	//i2c_smbus_write_byte_data(dsi85_i2c_client, DSI85_PLL_EN, 0x01);
	return;
}
#endif

#if 0
static int __init dsi85_init(void)
{
	int r = 0;

	r = i2c_add_driver(&dsi85_i2c_driver);
	if (r) {
		printk(KERN_WARNING "dsi85_i2c_driver" \
			" registration failed\n");
		return r;
	} else
		printk("DSI85-I2C: registered!\n");

	
	return 0;
}

static void __exit dsi85_exit(void)
{
	i2c_del_driver(&dsi85_i2c_driver);
}

module_init(dsi85_init);
module_exit(dsi85_exit);
#endif

module_i2c_driver(dsi85_i2c_driver);
MODULE_AUTHOR("NXP Semiconductor");
MODULE_DESCRIPTION("SN65DSI8X MIPI DSI to LVDS  driver");
MODULE_LICENSE("GPL");
